#include "ClassNameManager.h"
#include <Windows.h>
#include <tchar.h>
#include <assert.h>
#include "../MinHook/include/MinHook.h"

#ifdef _M_X64
#pragma comment(lib, "../lib/MinHook/MinHook.x64.lib")
#else
#pragma comment(lib,"../lib/MinHook/MinHook.x86.lib")
#endif

namespace Starts2000 {
	namespace Window {
		namespace Forms {

			#define FORM_CLASS_NAME L"WindowsForms10.Window.8.app"
			#define FORM_CUSTOM_CLASS_NAME L"Starts2000.Window"

			typedef ATOM (WINAPI * TrueRegisterClassW)(_In_ CONST WNDCLASSW *);
			typedef BOOL (WINAPI * TrueUnregisterClassW)(_In_ LPCWSTR, _In_opt_ HINSTANCE);
			typedef BOOL (WINAPI * TrueGetClassInfoW)(
				_In_opt_ HINSTANCE,
				_In_ LPCWSTR,
				_Out_ LPWNDCLASSW);
			typedef HWND (WINAPI * TrueCreateWindowExW)(
				_In_ DWORD,
				_In_opt_ LPCWSTR,
				_In_opt_ LPCWSTR,
				_In_ DWORD,
				_In_ int,
				_In_ int,
				_In_ int,
				_In_ int,
				_In_opt_ HWND,
				_In_opt_ HMENU,
				_In_opt_ HINSTANCE,
				_In_opt_ LPVOID);

			TrueRegisterClassW _registerClassW = NULL;
			TrueUnregisterClassW _unregisterClassW = NULL;
			TrueGetClassInfoW _getClassInfoW = NULL;
			TrueCreateWindowExW _createWindowExW = NULL;

			ATOM WINAPI RegisterClassWD(_In_ CONST WNDCLASSW *lpWndClass) {
				if (_tcsstr(lpWndClass->lpszClassName, FORM_CLASS_NAME)) {
					WNDCLASSW wndClass;
					memcpy(&wndClass, lpWndClass, sizeof(WNDCLASSW));
					wndClass.lpszClassName = FORM_CUSTOM_CLASS_NAME;
					auto ret = _registerClassW(&wndClass);
					return ret;
				}

				return _registerClassW(lpWndClass);
			}

			BOOL WINAPI UnregisterClassWD(_In_ LPCWSTR lpClassName, _In_opt_ HINSTANCE hInstance) {
				if (_tcsstr(lpClassName, FORM_CLASS_NAME)) {
					return _unregisterClassW(FORM_CUSTOM_CLASS_NAME, hInstance);
				}

				return _unregisterClassW(lpClassName, hInstance);
			}

			BOOL WINAPI GetClassInfoWD(_In_opt_ HINSTANCE hInstance,
				_In_ LPCWSTR lpClassName,
				_Out_ LPWNDCLASSW lpWndClass) {
				if (_tcsstr(lpClassName, FORM_CLASS_NAME)) {
					return _getClassInfoW(hInstance, FORM_CUSTOM_CLASS_NAME, lpWndClass);
				}

				return _getClassInfoW(hInstance, lpClassName, lpWndClass);
			}

			HWND WINAPI CreateWindowExWD(
				_In_ DWORD dwExStyle,
				_In_opt_ LPCWSTR lpClassName,
				_In_opt_ LPCWSTR lpWindowName,
				_In_ DWORD dwStyle,
				_In_ int X,
				_In_ int Y,
				_In_ int nWidth,
				_In_ int nHeight,
				_In_opt_ HWND hWndParent,
				_In_opt_ HMENU hMenu,
				_In_opt_ HINSTANCE hInstance,
				_In_opt_ LPVOID lpParam) {
				if (_tcsstr(lpClassName, FORM_CLASS_NAME)) {
					auto hwnd = _createWindowExW(dwExStyle, FORM_CUSTOM_CLASS_NAME, lpWindowName, dwStyle,
						X, Y, nWidth, nHeight, hWndParent, hMenu, hInstance, lpParam);
					assert(hwnd);
					return hwnd;
				}

				return _createWindowExW(dwExStyle, lpClassName, lpWindowName, dwStyle,
					X, Y, nWidth, nHeight, hWndParent, hMenu, hInstance, lpParam);
			}

			ClassNameManager::ClassNameManager() {
				auto ret = MH_Initialize();
				assert(ret == MH_STATUS::MH_OK);

				ret = MH_CreateHookApi(L"User32.dll", 
					"RegisterClassW", &RegisterClassWD, reinterpret_cast<LPVOID*>(&_registerClassW));
				assert(ret == MH_STATUS::MH_OK);

				ret = MH_CreateHookApi(L"User32.dll",
					"UnregisterClassW", &UnregisterClassWD, reinterpret_cast<LPVOID*>(&_unregisterClassW));
				assert(ret == MH_STATUS::MH_OK);

				ret = MH_CreateHookApi(L"User32.dll",
					"GetClassInfoW", &GetClassInfoWD, reinterpret_cast<LPVOID*>(&_getClassInfoW));
				assert(ret == MH_STATUS::MH_OK);

				ret = MH_CreateHookApi(L"User32.dll",
					"CreateWindowExW", &CreateWindowExWD, reinterpret_cast<LPVOID*>(&_createWindowExW));
				assert(ret == MH_STATUS::MH_OK);

				ret = MH_EnableHook(MH_ALL_HOOKS);
				assert(ret == MH_STATUS::MH_OK);
			}

			ClassNameManager::~ClassNameManager() {
			}

			ClassNameManager::!ClassNameManager() {
				auto ret = MH_Uninitialize();
				assert(ret == MH_STATUS::MH_OK);
			}
		}
	}
}